Imports System.Data
Imports System.Data.SqlClient


Public Class clSQLCompare
	Const cstDivider As String = "-------------------------------------"
	Const REPORT_TITLE = "COPS DBMS Server Comparison Report"
	Public ReadOnly Property ReportTitle() As String
		Get
			Return REPORT_TITLE
		End Get
	End Property
	Public Shared Sub SQLCompareMain( _
	 ByVal strStdDatabase As String, _
	 ByVal strCompDatabase As String, _
	 ByRef strResults As String, _
	 ByRef strBody As String)
		', _
		' ByVal strStdObjects As clDBObjects, _
		' ByVal strCompObjects As clDBObject, _
		' ByVal strStdTblObjects As clDBTableObjects, _
		' ByVal strCompTblObjects As clDBTableObject)

		strResults = ""
		If Not ValidateInput(strStdDatabase) OrElse Not ValidateInput(strCompDatabase) Then
			WriteStatus(strResults, _
			 "Confirm that Standard and Comparison Object names are using ""databaseservername.databasename"" format", _
			 True)
			Exit Sub
		End If


		Dim connStdDBMS As New SqlConnection()							'Connection for standard objects
		Dim connCompDBMS As New SqlConnection()							'Connection for comparison objects

		Dim strStdObjects As New clDBObjects(strStdDatabase)							'Collection of standard objects
		Dim strCompObjects As New clDBObjects(strCompDatabase)							'Collection of comparison objects
		Dim strStdTblObjects As New clDBTableObjects(strStdDatabase)			'Collection of standard schema
		Dim strCompTblObjects As New clDBTableObjects(strCompDatabase)			'Collection of comparison schema



		Try

			Call CompareObjects(strResults, connStdDBMS, connCompDBMS, strStdObjects, strCompObjects)

			Call CompareSchemas(strResults, connStdDBMS, connCompDBMS, strStdObjects, strCompObjects, _
			strStdTblObjects, strCompTblObjects)

			'Exclude DDL match until it can be improved.
			'Call CompareDDL(strResults, strStdObjects, strCompObjects)

			WriteStatus(strResults, "Finished comparing objects in the databases", True, True)
			Call ReportResults(strResults, strStdObjects, strCompObjects, strStdTblObjects, strCompTblObjects, strBody)

		Catch ex As Exception
			WriteStatus(strResults, Err.Description, True, True)
			WriteStatus(strResults, "Comparison ended unsuccessfully")
		End Try

		connStdDBMS.Close()
		connCompDBMS.Close()


	End Sub
	Public Shared Sub CompareObjects(ByRef strResults As String, ByRef connStdDBMS As SqlConnection, ByRef connCompDBMS As SqlConnection, ByRef strStdObjects As clDBObjects, ByRef strCompObjects As clDBObjects)

		Dim cmdSQL As New SqlCommand()											'Command object for SQl text
		Dim rdSQLReader As SqlDataReader										'Reader for connection object
		Dim strSQL As String = ""														'Contains SQL text to be run on remote DBMSs

		Dim aDBObject As New clDBObject()										'DBObject container


		WriteStatus(strResults, "Checking to determine if the databases are accessible...", True, True)

		'Determine if we can reach the standard objects that were specified
		Call ConnectDBMS(connStdDBMS, strStdObjects.Database_Name, strResults)

		'If we failed, close gracefully.. On Error statement will direct us to the end
		If strResults.Contains("not found") Then
			Throw New System.Exception("Error thrown while accessing standard database.")
			Exit Sub
		End If

		'Determine if we can reach the comparison objects that were specified
		Call ConnectDBMS(connCompDBMS, strCompObjects.Database_Name, strResults)

		'If we failed, close gracefully.. On Error statement will direct us to the end
		If strResults.Contains("not found") Then
			Throw New System.Exception("Error thrown while accessing comparison database.")
			Exit Sub
		End If

		'Write a standard header to the status log
		WriteStatus(strResults, "Comparing objects in the databases...", True, True, True)

		'Store the contents of the "SysobjectsNameAndType.sql" file for processing
		strSQL = My.Computer.FileSystem.ReadAllText(AppDomain.CurrentDomain.SetupInformation.ApplicationBase & "\SysobjectsNameAndType.sql")

		'Query the standard DBMS
		cmdSQL.CommandText = strSQL
		cmdSQL.Connection = connStdDBMS
		rdSQLReader = cmdSQL.ExecuteReader()

		'Walk through each row of the data reader collecting the result set into an array.
		Do While rdSQLReader.Read()
			aDBObject = New clDBObject
			aDBObject.Object_Name = rdSQLReader("ObjectName")
			aDBObject.SaveObjectType = rdSQLReader("ObjectType")
			aDBObject.Object_ID = rdSQLReader("ObjectID")
			aDBObject.Object_DDL = rdSQLReader("ObjectDDL")
			aDBObject.DDL_Soundex = rdSQLReader("Soundex")
			strStdObjects.Add(aDBObject)
		Loop
		rdSQLReader.Close()

		'Query the comparison DBMS
		cmdSQL.Connection = connCompDBMS
		rdSQLReader = cmdSQL.ExecuteReader()

		'Walk through each row of the data reader collecting the result set into an array.
		Do While rdSQLReader.Read()
			aDBObject = New clDBObject
			aDBObject.Object_Name = rdSQLReader("ObjectName")
			aDBObject.SaveObjectType = rdSQLReader("ObjectType")
			aDBObject.Object_ID = rdSQLReader("ObjectID")
			aDBObject.Object_DDL = rdSQLReader("ObjectDDL")
			aDBObject.DDL_Soundex = rdSQLReader("Soundex")
			strCompObjects.Add(aDBObject)
		Loop
		rdSQLReader.Close()


		'Find objects in the comparison database that are not in the standard database
		'Dim strStdObject, strCompObject As clDBObject
		'Compare each standard object to every comparison object. If a match is found, save it
		For Each strStdObject As clDBObject In strStdObjects
			For Each strCompObject As clDBObject In strCompObjects
				'If the objects match, save the data and prepare to go to the next standard object
				If [String].Compare(strCompObject.Object_Name, strStdObject.Object_Name) = 0 _
				 And strCompObject.Object_Type.CompareTo(strStdObject.Object_Type) = 0 Then
					strStdObject.NameTypeMatched = True
					strCompObject.NameTypeMatched = True
					Exit For
				End If
			Next strCompObject
			'If no matching standard object was found, save it and report it as missing to the results collection
			If Not strStdObject.NameTypeMatched Then
				With strStdObject
					WriteStatus(strResults, _
					 .Object_Name & "{" & .DisplayObjectType & "} exists in " & strStdObjects.Database_Name & _
					 " but not in " & strCompObjects.Database_Name)
				End With
			End If
		Next strStdObject

		WriteStatus(strResults, "", False, True)

		'Find objects in the standard database that are not in the comparison database
		For Each strCompObject As clDBObject In strCompObjects
			'If no matching comparison object was found, save it and report it as missing to the results collection
			If Not strCompObject.NameTypeMatched Then
				With strCompObject
					WriteStatus(strResults, _
					.Object_Name & "{" & .DisplayObjectType & "} exists in " & strCompObjects.Database_Name & _
					" but not in " & strStdObjects.Database_Name)
				End With
			End If

		Next strCompObject

	End Sub



	Public Shared Sub CompareSchemas(ByRef strResults As String, ByRef connStdDBMS As SqlConnection, ByRef connCompDBMS As SqlConnection, ByRef strStdObjects As clDBObjects, ByRef strCompObjects As clDBObjects, ByRef strStdTblObjects As clDBTableObjects, ByRef strCompTblObjects As clDBTableObjects)


		'If there are no matches there is no need to run
		If strStdObjects.MatchedObjectCount = 0 Then
			Exit Sub
		End If

		Dim cmdSQL As New SqlCommand()											'Command object for SQL text
		Dim rdSQLReader As SqlDataReader										'Reader for connection object
		Dim strSQL As String = ""														'Contains SQL text to be run on remote DBMSs

		Dim aDBObject As New clDBTableObject								'DBTableObject container



		'Write a standard header to the status log
		WriteStatus(strResults, "Comparing schemas for objects that match in the databases...", True, True, True)

		'Modify the SQL string to include an IN clause containing the matching values
		strSQL = _
		My.Computer.FileSystem.ReadAllText(AppDomain.CurrentDomain.SetupInformation.ApplicationBase & "\SysobjectsColumnsAndLength.sql")

		strSQL = Replace(strSQL, "IN CLAUSE GETS INSERTED HERE", _
		strStdObjects.MatchedIDsINClause, , , CompareMethod.Text)

		'Query the standard DBMS
		cmdSQL.CommandText = strSQL
		cmdSQL.Connection = connStdDBMS
		rdSQLReader = cmdSQL.ExecuteReader()

		'Walk through each row of the data reader collecting the result set
		Do While rdSQLReader.Read()
			aDBObject = New clDBTableObject
			aDBObject.Object_Name = rdSQLReader("ObjName")
			aDBObject.SaveObjectType = rdSQLReader("ObjType")
			aDBObject.Column_Name = rdSQLReader("ColName")
			aDBObject.Data_Type = rdSQLReader("DataType")
			aDBObject.Column_Length = rdSQLReader("ColLngth")
			aDBObject.Object_ID = rdSQLReader("ObjID")
			strStdTblObjects.Add(aDBObject)
		Loop
		rdSQLReader.Close()

		'Modify the SQL string to include an IN clause containing the matching values
		strSQL = Replace(strSQL, "IN CLAUSE GETS INSERTED HERE", _
		strCompObjects.MatchedIDsINClause, , , CompareMethod.Text)

		'Query the comparison DBMS
		cmdSQL.CommandText = strSQL
		cmdSQL.Connection = connCompDBMS
		rdSQLReader = cmdSQL.ExecuteReader()

		'Walk through each row of the data reader collecting the result set.
		Do While rdSQLReader.Read()
			aDBObject = New clDBTableObject
			aDBObject.Object_Name = rdSQLReader("ObjName")
			aDBObject.SaveObjectType = rdSQLReader("ObjType")
			aDBObject.Column_Name = rdSQLReader("ColName")
			aDBObject.Data_Type = rdSQLReader("DataType")
			aDBObject.Column_Length = rdSQLReader("ColLngth")
			aDBObject.Object_ID = rdSQLReader("ObjID")
			strCompTblObjects.Add(aDBObject)
		Loop
		rdSQLReader.Close()



		'Find objects in the comparison database that have different datatype specs in the standard database
		'Compare each standard object to every comparison object. If a match is found, save it
		For Each strStdTblObject As clDBTableObject In strStdTblObjects
			For Each strCompTblObject As clDBTableObject In strCompTblObjects
				'If the objects match, save the data and prepare to go to the next standard object
				If [String].Compare(strCompTblObject.Object_Name, strStdTblObject.Object_Name) = 0 _
				And strCompTblObject.Object_Type.CompareTo(strStdTblObject.Object_Type) = 0 Then
					If [String].Compare(strCompTblObject.Column_Name, strStdTblObject.Column_Name) = 0 Then
						strCompTblObject.ColumnNameMatched = True
						strStdTblObject.ColumnNameMatched = True
						If [String].Compare(strCompTblObject.Ext_Data_Type, strStdTblObject.Ext_Data_Type) = 0 Then
							strCompTblObject.DataTypeMatched = True
							strStdTblObject.DataTypeMatched = True
							Exit For
						End If
					End If
				End If
			Next strCompTblObject
			'If no matching standard object was found, save it and report it as missing to the results collection
			If Not strStdTblObject.DataTypeMatched Then
				WriteStatus(strResults, _
				 strStdTblObject.FullName & " has different datatype specifications.")
			ElseIf Not strStdTblObject.ColumnNameMatched Then
				WriteStatus(strResults, _
				 strStdTblObject.FullName & " exists in " & strStdTblObjects.Database_Name & _
				 " but not in " & strCompTblObjects.Database_Name)
			End If
		Next strStdTblObject

		WriteStatus(strResults, "", False, True)

		'No need to do this because we matched the tablespaces in COmpareObjects above.
		''Find objects in the standard database that are not in the comparison database
		'For Each strCompTblObject As clDBTableObject In strCompTblObjects
		'	'If no matching comparison object was found, save it and report it as missing to the results collection
		'	If Not strCompTblObject.DataTypeMatched Then
		'		WriteStatus(strResults, _
		'		 strCompTblObject.FullName & " has different datatype specifications.")
		'	ElseIf Not strCompTblObject.ColumnNameMatched Then
		'		WriteStatus(strResults, _
		'		 strCompTblObject.FullName & " exists in " & strCompTblObjects.Database_Name & _
		'		 " but not in " & strStdObjects.Database_Name)
		'	End If
		'Next strCompTblObject



	End Sub
	Public Shared Sub CompareDDL(ByRef strResults As String, ByRef strStdObjects As clDBObjects, ByRef strCompObjects As clDBObjects)
		'Compare the DDL for non table objects in this sub

		'Write a standard header to the status log
		WriteStatus(strResults, "Comparing DDL for objects that match in the databases...", True, True, True)

		For Each strStdObject As clDBObject In strStdObjects
			If strStdObject.NameTypeMatched Then
				For Each strCompObject As clDBObject In strCompObjects
					'If the objects match, the class wull update the values and we go to the next standard object
					If Not strCompObject.IsTableObject AndAlso strCompObject.IsDDLMatched(strStdObject) Then
						Exit For
					End If
				Next strCompObject
				'If they do not match
				If Not strStdObject.DDLMatched Then
					With strStdObject
						WriteStatus(strResults, _
						 "DDL in " & .DisplayObjectType & " " & strStdObjects.Database_Name & "." & .Object_Name & _
						 " does not match the DDL of the same object in " & strCompObjects.Database_Name)
					End With
				End If
			End If
		Next strStdObject
	End Sub

	Public Shared Sub ConnectDBMS(ByVal cnnConnection As SqlConnection, ByVal strDatabase As String, ByRef strResults As String)
		Dim strObject As String()						'Array containing DBMS and database names

		'Determine if we can reach the objects that were specified
		Try
			'Split out DBMS and database name into array using period as a delimiter
			strObject = strDatabase.Split(".")

			'Set the connection string
			cnnConnection.ConnectionString = "Integrated Security=SSPI;" & _
			"Persist Security Info=False;Initial Catalog=" & strObject(1) & ";Data Source=" & _
			strObject(0)

			'Open the connection
			cnnConnection.Open()

			'Handle connection errors here
		Catch Ex As Exception
			WriteStatus(strResults, strDatabase & " not found. " & Ex.Message)
			Exit Sub
		End Try

		WriteStatus(strResults, strDatabase & " found.")

	End Sub

	Public Shared Sub WriteStatus(ByRef strMessageVariable As String, _
	 ByVal strStatusMessage As String, _
	 Optional ByVal boolUseTimeStamp As Boolean = False, _
	 Optional ByVal boolUseSeparator As Boolean = False, _
	 Optional ByVal boolInsertLine As Boolean = False)

		'Universal log writer with optional timestamp and divider bars 

		'If the message is to be preceeded by a line, insert it before the message
		If boolInsertLine Then
			strMessageVariable &= vbCrLf
		End If

		'If this message is to be timestamped, insert it before the text
		If boolUseTimeStamp Then
			strMessageVariable &= Now.ToString & " "
		End If

		'Insert the text in all conditions
		strMessageVariable &= strStatusMessage & vbCrLf

		'If this message is to be followed by a separator, insert it after the message
		If boolUseSeparator Then
			strMessageVariable &= cstDivider & vbCrLf
		End If


	End Sub
	Public Shared Function ValidateInput(ByVal txtInputString As String) As Boolean
		If txtInputString Like "*?.?*" Then
			Return True
		Else
			Return False
		End If
	End Function
	Public Shared Sub ReportResults(ByRef strResults As String, ByRef strStdObjects As clDBObjects, ByRef strCompObjects As clDBObjects, ByRef strStdTblObjects As clDBTableObjects, ByRef strCompTblObjects As clDBTableObjects, ByRef strBody As String)

		WriteStatus(strResults, "Preparing detailed report", , , True)

		'Write the report headings on the report page variable using HTML
		'strBody = ""
		'strBody &= "<!DOCTYPE HTML PUBLIC ""-//W3C//DTD HTML 4.01 Transitional//EN"">" & vbNewLine
		'strBody &= "<html>" & vbNewLine & vbTab & "<head><title>" & REPORT_TITLE & "</title></head>" & vbNewLine & vbTab
		'strBody &= "<body>"
		strBody &= "<center><h1>" & REPORT_TITLE & "</h1></center>" & vbNewLine & vbTab

		strBody &= "<center><h3>Database Object Mismatches</h3></center>" & vbNewLine & vbTab
		strBody &= "<table width='100%'>" & vbNewLine & vbTab
		strBody &= "<tr>"
		strBody &= "<td width='2%'></td>"
		strBody &= "<td width='49%'><u>Standard Database</u></td>"
		strBody &= "<td width='49%'><u>Comparison Database</u></td>"
		strBody &= "</tr>" & vbNewLine & vbTab

		strBody &= "<tr>"
		strBody &= "<td width='2%'></td>"
		strBody &= "<td width='49%'><font color='blue'>" & strStdObjects.Database_Name & "</font></td>"
		strBody &= "<td width='49%'><font color='red'>" & strCompObjects.Database_Name & "</font></td>"
		strBody &= "</tr>" & vbNewLine & vbTab & vbNewLine & vbTab

		strBody &= "<tr>"
		strBody &= "<td width='2%'></td>"
		strBody &= "<td width='49%'><u>Missing Objects</u></td>"
		strBody &= "<td width='49%'><u>Missing Objects</u></td>"
		strBody &= "</tr>" & vbNewLine & vbTab

		Dim strHTMLTable(1000, 2) As String
		Dim i As Integer = 0

		For Each strStdObject As clDBObject In strStdObjects
			If Not strStdObject.NameTypeMatched Then
				strHTMLTable(i, 0) = strStdObject.Object_Name & "{" & strStdObject.DisplayObjectType & "}"
				i += 1
			End If
		Next strStdObject

		i = 0
		For Each strCompObject As clDBObject In strCompObjects
			If Not strCompObject.NameTypeMatched Then
				strHTMLTable(i, 1) = strCompObject.Object_Name & "{" & strCompObject.DisplayObjectType & "}"
				i += 1
			End If
		Next strCompObject

		For j As Integer = 0 To 1000
			If [String].IsNullOrEmpty(strHTMLTable(j, 0)) AndAlso [String].IsNullOrEmpty(strHTMLTable(j, 1)) Then
				Exit For
			Else
				strBody &= "<tr>"
				strBody &= "<td width='2%'></td>"
				strBody &= "<td width='49%'>" & strHTMLTable(j, 1) & "</td>"
				strBody &= "<td width='49%'>" & strHTMLTable(j, 0) & "</td>"
				strBody &= "</tr>" & vbNewLine & vbTab
			End If
		Next j
		strBody &= "</table><br/>" & vbNewLine & vbTab & vbNewLine & vbTab


		strBody &= "<center><h3>Data Type Specification Differences (Tablespace Objects Only)</h3></center>" & vbNewLine & vbTab
		strBody &= "<table width='100%'>" & vbNewLine & vbTab
		strBody &= "<tr>"
		strBody &= "<td width='60%'><u>Tablespace Name</u></td>"
		strBody &= "<td width='20%'><u>Standard Database</u></td>"
		strBody &= "<td width='20%'><u>Comparison Database</u></td>"
		strBody &= "</tr>" & vbNewLine & vbTab


		For Each strStdTblObject As clDBTableObject In strStdTblObjects
			For Each strCompTblObject As clDBTableObject In strCompTblObjects
				If strStdTblObject.ColumnNameMatched AndAlso [String].Compare(strStdTblObject.Column_Name, strCompTblObject.Column_Name) Then
					If Not strStdTblObject.DataTypeMatched Then
						strBody &= "<tr>"
						strBody &= "<td width='60%'>" & strStdTblObject.Object_Name & "." & strStdTblObject.Column_Name & "</td>"
						strBody &= "<td width='20%'>" & strStdTblObject.Ext_Data_Type & "</td>"
						strBody &= "<td width='20%'>" & strCompTblObject.Ext_Data_Type & "</td>"
						strBody &= "</tr>" & vbNewLine & vbTab
						Exit For
					End If
				End If
			Next strCompTblObject
			If Not strStdTblObject.ColumnNameMatched Then
				strBody &= "<tr>"
				strBody &= "<td width='60%'>" & strStdTblObject.FullName & "</td>"
				strBody &= "<td width='20%'>" & strStdTblObject.Ext_Data_Type & "</td>"
				strBody &= "<td width='20%'>" & "Column Does Not Exist" & "</td>"
				strBody &= "</tr>" & vbNewLine & vbTab
			End If
		Next strStdTblObject

		For Each strCompTblObject As clDBTableObject In strCompTblObjects
			If Not strCompTblObject.ColumnNameMatched Then
				strBody &= "<tr>"
				strBody &= "<td width='60%'>" & strCompTblObject.FullName & "</td>"
				strBody &= "<td width='20%'>" & "Column Does Not Exist" & "</td>"
				strBody &= "<td width='20%'>" & strCompTblObject.Ext_Data_Type & "</td>"
				strBody &= "</tr>" & vbNewLine & vbTab
			End If
		Next strCompTblObject


		strBody &= "</table><br/>" & vbNewLine & vbTab & vbNewLine & vbTab

		'DDL reporting removed until repaired
		'strBody &= vbNewLine & vbNewLine
		'strBody &= "<center><h3>DDL Discrepancies (Non-Tablespace Objects Only)</h3></center>" & vbNewLine & vbTab

		'For Each strStdObject As clDBObject In strStdObjects
		'	If strStdObject.NameTypeMatched AndAlso Not strStdObject.DDLMatched AndAlso Not strStdObject.IsTableObject Then
		'		strBody &= strStdObject.Object_Name & "{" & strStdObject.DisplayObjectType & "}" & "</br>"
		'	End If
		'Next

		strBody &= vbNewLine & vbTab & "<center><small>Report completed on " & CType(Now(), String)
		strBody &= "<br/> 2007 AT&T Knowledge Ventures, L.P. All rights reserved.</small></center>" & vbNewLine & vbTab
		'strBody &= "</body>" & vbNewLine & "</html>"

		WriteStatus(strResults, "Finished preparing detailed report", True, True)

	End Sub



End Class